﻿using Sandbox;
using Sandbox.Common;
using Sandbox.Common.Components;
using Sandbox.Common.ObjectBuilders;
using Sandbox.ModAPI;
using Sandbox.ModAPI.Ingame;
using System.Collections.Generic;
using VRageMath;

/// <summary>
/// version 0.6:
///     Update API and code to work with version 01.072
///     
///
/// Version 0.5:
///     Refactored code to a core seperate from any individual block
///     Teleporter now also prints it's removal when block is removed
///     
/// 
/// Version 0.4: 
///     fixed offset and rotation (almost perfect), 
///     added 3 cooldown for destination teleporter, 
///     added more tags: "teleporter", "teleport", "tp"
/// Version 0.3: Refactored code in hopes of losing the compile errors
/// Version 0.2: Cleaned up debug texts, increased range 7
/// Version 0.1: Initial release
/// </summary>
namespace TeleporterMechanisms {
    public class TeleporterCore : MyGameLogicComponent {
        //squared so we can use the slightly faster squared distance function
        float maxDistance = 49;
        float maxPilotedDistance = 1000;

        //tags for the teleporter
        static string[] TAGS = { "teleporter", "teleport", "tp" };
        static string[] GATETAGS = { "gate" };
        int tagMatched = -1;
        bool piloted = false;

        //cooldown to set on the destination
        static int cooldownDuration = 3;
        int cooldown = 0;

        //do we need to power down?
        private bool m_disable = false;

        //static list of teleporters
        public static List<TeleporterCore> teleporters = new List<TeleporterCore>();
        public static List<TeleporterCore> gates = new List<TeleporterCore>();

        //reference to the functional block (any block with power on/off)
        Sandbox.ModAPI.IMyFunctionalBlock functional;
        MyEntityUpdateEnum originalUpdates;

        public string Name {
            get { return functional.CustomName; }
        }

        public override void Init( MyObjectBuilder_EntityBase objectBuilder ) {
            //originalUpdates = Entity.NeedsUpdate;
            //if ((Entity as IMyBeacon).BlockDefinition.SubtypeId.Contains("Beacon"))
            //return;
            
            base.Init(objectBuilder);
            Entity.NeedsUpdate |= MyEntityUpdateEnum.EACH_FRAME;
        }
        
        public override MyObjectBuilder_EntityBase GetObjectBuilder(bool copy = false)//fuzz edit
        {
            return Entity.GetObjectBuilder();
        }
        
        private void _Init() {
            functional = Entity as Sandbox.ModAPI.IMyFunctionalBlock;

            //For some reason the Entity cast does not play well with Init
            functional.EnabledChanged += OnEnableChanged;

            //DebugPrint( "Info: " + functional.DetailedInfo );
            //DebugPrint( Name + " range: " + functional.WorldVolume.Radius );

            //this causes SE to chrash: Delegates must be of the same type ???
            //functional.CustomNameChanged += OnNameChange;
        }

        public override void MarkForClose() {
            base.MarkForClose();

            //functional.EnabledChanged -= OnEnableChanged;
            //functional.CustomNameChanged -= OnNameChange;

            UnregisterTeleporter();
        }

        public override void UpdateAfterSimulation() {
            //Init hack
            if( functional == null ) {
                //Remove each frame update, change to each 100th frame
                Entity.NeedsUpdate = originalUpdates | MyEntityUpdateEnum.EACH_100TH_FRAME;

                _Init();
            }
        }

        public override void UpdateAfterSimulation100() {
            if( functional == null ) {
                _Init();
            } else {
                //sadly, can't use functional.CustomNameChanged, so we check each 100th update :-(
                NameCheck();

                if( cooldown > 0 ) {
                    cooldown--;
                }

                if( m_disable ) {
                    functional.RequestEnable( false );
                    m_disable = false;
                }
            }
        }


        public void OnNameChange( Sandbox.ModAPI.IMyTerminalBlock obj ) {
            NameCheck();
        }

        public void OnEnableChanged( Sandbox.ModAPI.IMyTerminalBlock obj ) {
            if( functional == null ) {
                //DebugPrint( "Teleporter OnEnableChanged error; functional is null" );
                return;
            }

            if( cooldown > 0 ) {
                //DebugPrint( "Teleporter cooling down: " + cooldown, 2000 );
                return;
            }

            if( functional.Enabled ) {
                if( NameCheck() ) {
                    //DebugPrint( "Teleporter activating: " + Name );

                    //tried using these before, with SE chrashing
                    //functional.PersistentFlags &= ~MyPersistentEntityFlags2.Enabled;
                    //functional.RequestEnable( false );
                    m_disable = true;

                    var useList = piloted ? gates : teleporters;

                    Vector3 checkSpot = GetOriginPosition();

                    //find the destination
                    Vector3 teleSpot;
                    Matrix worldMatrix;
                    TeleporterCore other = null;
                    {
                        string name = Lower( Name );
                        string tag = (piloted ? GATETAGS : TAGS)[tagMatched];
                        string match = name.Substring( tag.Length );

                        var word = (piloted ? "Gates" : "Teleporters");

                        List<TeleporterCore> list = new List<TeleporterCore>();
                        if( match.Length > 0 ) {
                            list = useList.FindAll( tp => Lower( tp.Name.Substring( tag.Length ) ) == match );
                        } else {
                            list = useList.FindAll( tp => Lower( tp.Name ) == tag );
                        }
                        //DebugPrint( word + " matching '" + match + "': " + list.Count );

                        if( list.Count < 2 ) {
                            //DebugPrint( "You need at least two " + word + " named: " + Name );
                            //DebugPrint( word + " known: " + useList.Count );
                            //DebugPrint( word + " matching: " + list.Count );
                            return;
                        }

                        int id = list.IndexOf( this );
                        id++;
                        if( id >= list.Count )
                            id = 0;

                        other = list[id];
                        worldMatrix = other.Entity.WorldMatrix;
                    }

                    //find the player to teleport
                    {
                        // First attempt: The hard way of doing things
                        //
                        //List<IMyPlayer> players = new List<IMyPlayer>();
                        //Func<IMyPlayer, bool> func = x => {
                        //    Vector3 pos = x.PlayerCharacter.Entity.GetPosition();
                        //    return Vector3.Distance( pos, checkSpot ) <= maxDistance;
                        //};
                        //MyAPIGateway.Players.GetPlayers( players, func );

                        //if( players.Count > 0 ) {
                            //start teleporting the players
                        //    foreach( IMyPlayer player in players ) {
                        //        Vector3 pos = player.PlayerCharacter.Entity.GetPosition();
                        //        Vector3 offset = checkSpot - pos;
                        //        player.PlayerCharacter.Entity.SetPosition( teleSpot + offset );
//
                        //        DebugPrint( "I would like to teleport " + player.DisplayName + "now" );
                        //    }
                        //}

                        // this however might just be enough, not sure how this works with multiplayers
                        IMyEntity player;
                        if( piloted ) {
                            var ce = MyAPIGateway.Session.Player.Controller.ControlledEntity;
                            if( ce == null ) {
                                //DebugPrint( "Not flying anything" );
                                return;
                            }
                            player = ce.Entity.GetTopMostParent();
                        } else {
                            //player = MyAPIGateway.Session.Player.PlayerCharacter.Entity;
                            player = MyAPIGateway.Session.Player.Controller.ControlledEntity.Entity;
                        }
                        Vector3 diff = checkSpot - player.GetPosition();

                        bool closeEnough = diff.LengthSquared() < (piloted ? maxPilotedDistance : maxDistance);

                        //square the distance, saves the root operator
                        if( closeEnough ) {
                            //Vector3 pos = diff
                            var m = other.GetDestinationMatrix( diff, player, this );
                            player.WorldMatrix = m;
                            if( piloted ) {
                                //DebugPrint( "Gated " + player.DisplayName + " to gate at: " + other.GetOriginPosition() );
                            } else {
                                //DebugPrint( "Teleported " + player.DisplayName + " to teleporter at: " + other.GetOriginPosition() );
                            }
                            other.cooldown = cooldownDuration;
                        }
                    }
                }
            }
        }

        protected bool NameCheck() {
            //Check if the CustomName matches the defined TAG, if so return true
            string name = Lower( Name );

            //we also need to be sure the teleporter list remains accurate
            //this should remain OK as long as you don't use thousands of teleporters
            for( int i=0; i < TAGS.Length; i++ ) {
                string t = TAGS[i];
                if( name.StartsWith( t ) ) {
                    if( piloted )
                        UnregisterTeleporter();
                    piloted = false;
                    tagMatched = i;
                    RegisterTeleporter();
                    return true;
                }
            }

            for( int i=0; i < GATETAGS.Length; i++ ) {
                string t = GATETAGS[i];
                if( name.StartsWith( t ) ) {
                    if( !piloted )
                        UnregisterTeleporter();
                    piloted = true;
                    tagMatched = i;
                    RegisterTeleporter();
                    return true;
                }
            }

            UnregisterTeleporter();
            return false;
        }

        private void RegisterTeleporter() {
            var uselist = piloted ? gates : teleporters;
            var word = piloted ? "Gate" : "Teleporter";

            if( !uselist.Contains( this ) ) {
                uselist.Add( this );
                //DebugPrint( word + " registered: " + Name );
            }
        }

        private void UnregisterTeleporter() {
            var uselist = piloted ? gates : teleporters;
            var word = piloted ? "Gate" : "Teleporter";

            if( uselist.Contains( this ) ) {
                uselist.Remove( this );
                //DebugPrint( word + " removed (name changed): " + Name );
            }
        }

        protected virtual Vector3 GetOriginPosition() {
            return functional.GetPosition();
        }

        protected virtual Matrix GetDestinationMatrix( Vector3 offset, IMyEntity player, TeleporterCore source ) {
            var teleSpot = functional.GetPosition();
            //lets only change the UP direction and hope the matrix will sort it out
            return Matrix.CreateWorld( teleSpot - offset, player.WorldMatrix.Forward, source.Entity.WorldMatrix.Up );
        }

        public static string Lower( string s ) {
            return s.ToLowerInvariant();
        }

        public static void DebugPrint( string text, int duration = 5000 ) {
            MyAPIGateway.Utilities.ShowNotification( text, duration, MyFontEnum.Red );
        }
    }
}